{ Unit:            ADJ_LIST.pas
  Autor:           Sascha Nemecek
  Matrikelnr.:     9825815
  Erstellt am:     17.05.1999
  Letzte nderung: 04.06.1999

  Kurzbeschreibung:
  }


UNIT adj_list;

INTERFACE

USES linlist2;


TYPE
   string4 = string [4];
   string20 = string [20];

   { Definition eines AdjazenzKnotens. }
   AdjKnotenPtr = ^AdjazenzKnoten;
   AdjazenzKnoten = RECORD
      kurzbez:         string4;
      name:            string20;
      min, max:        integer;
      besucht:         boolean;
      partOf, parts:   ListePtr;
   END;


   { Definition der Klasse: AdjazenzListe. }
   AdjPtr = ^AdjazenzListe;
   AdjazenzListe = OBJECT
{      PRIVATE}
         adj:   ListePtr;

{      PUBLIC}
         CONSTRUCTOR init;
         DESTRUCTOR done;

         FUNCTION add_Knoten (bez: string4; name: string20;
                          min, max: integer): boolean;
         FUNCTION add_Link (source, dest: string4) : boolean;
         FUNCTION chkrek (h: AdjKnotenPtr) : boolean;
         FUNCTION find (s: string4) : AdjKnotenPtr;
         FUNCTION get_alle_Knoten   : ListePtr;
         FUNCTION isEmpty : boolean;
         FUNCTION linearisieren (bez : string4; var h : ListePtr;
                                var min, max : integer) : boolean;
         FUNCTION remove_Knoten (bez : string4) : boolean;
         FUNCTION remove_Link (source, dest : string4) : boolean;
         FUNCTION start_Knoten : ListePtr;
         FUNCTION ziel_Knoten  : ListePtr;

         PROCEDURE kritischer_Pfad (bez: string4; var max: integer;
                                var krit: ListePtr);
         PROCEDURE reset_Besucht;
   END;




IMPLEMENTATION


{ Konstruktor.
  Initialisiert die Liste der Adjazenzknoten. }
CONSTRUCTOR AdjazenzListe.init;
BEGIN
   adj := new (ListePtr, Init);
END;




{ Destructor.
  Lscht die Liste der Adjazenzknoten. }
DESTRUCTOR AdjazenzListe.done;
BEGIN
   dispose (adj, done);
END;




{ Funktion add_Knoten
  Fgt einen neuen Knoten in die Adjazenzliste ein.
  @param  bez   Kurzbezeichnung des Prozesses.
  @param  name  Name des Prozesses.
  @param  min   Minimale Dauer des Prozesses.
  @param  max   Maximale Dauer des Prozesses.
  @return TRUE  Wenn der Knoten eingefgt werden konnte.
          FALSE Wenn der Knoten nicht eingefgt wurde. }
FUNCTION AdjazenzListe.add_Knoten
         (bez: string4; name: string20; min, max: integer) : boolean;

VAR
   h: AdjKnotenPtr;

BEGIN
   { Wenn Prozess schon vorhanden, Abbruch. }
   IF find (bez) <> NIL THEN
      add_Knoten := FALSE

   ELSE BEGIN
      { Knoten erzeugen. }
      h          := new (AdjKnotenPtr);

      { Variable initialisieren. }
      h^.kurzbez := bez;
      h^.name    := name;
      h^.min     := min;
      h^.max     := max;
      h^.besucht := false;
      h^.partOf  := new (ListePtr, Init);
      h^.parts   := new (ListePtr, Init);

      adj^.addElement (h);

      add_Knoten := TRUE;
   END;
END;




{ Funktion add_Link.
  Fgt zwischen zwei Knoten eine Verknpfung ein.
  @param  source Der bergeordnete Knoten.
  @param  dest   Der untergeordente Knoten.
  @return TRUE   Wenn die Verknpfung eingefgt werden konnte.
          FALSE  Wenn die Verknpfung nicht eingefgt wurde. }
FUNCTION AdjazenzListe.add_Link (source, dest: string4) : boolean;

VAR
   h1, h2: AdjKnotenPtr;
   ok    : boolean;

BEGIN
   { Knoten suchen. }
   h1 := find (source);
   h2 := find (dest);
   ok := TRUE;

   { Wenn einer nicht gefunden werden konnte - ABBRUCH. }
   IF (h1 = NIL) OR (h2 = NIL) THEN
      add_Link := FALSE

   ELSE BEGIN

      { berprfung, ob diese Verbindung schon existiert. }
      h1^.Parts^.gotoStart;
      WHILE (NOT h1^.parts^.eol) AND ok DO BEGIN
         IF h1^.Parts^.getData = h2 THEN
            ok := FALSE;
         h1^.Parts^.gotoNext;
      END;

      IF ok THEN BEGIN
         { Verknpfung einfgen. }
         h2^.partOf^.addElement (h1);
         h1^.parts^.addElement (h2);

         { Wenn dadurch eine Rekursion entsteht, wieder entfernen. }
         IF chkrek (h1) = TRUE THEN BEGIN
            ok := remove_Link (source, dest);
            add_Link := FALSE;
         END ELSE
            add_Link := TRUE;

      { Wenn Abhngigkeit bereits existiert, FALSE. }
      END ELSE
         add_Link := FALSE;
   END;
END;




{ Funktion chkrek.
  berprft, ob der Graph rekursiv ist.
  @param  h Der Knoten, der gerade untersucht wird.
            Aufruf mit NIL untersucht den ganzen Graphen.
  @return TRUE  der Graph ist rekursiv.
          FALSE der Graph ist rekursionsfrei. }
FUNCTION AdjazenzListe.chkrek (h: AdjKnotenPtr) : boolean;

VAR
   t  : AdjKnotenPtr;
   rek: boolean;

BEGIN
   rek := FALSE;

   { Wenn kein Wert bergeben, initialisieren. }
   IF h = NIL THEN BEGIN
      adj^.gotostart;
      h := adj^.getData;
      h^.besucht := TRUE;

   END ELSE
      { Rekursionsabfrage. }
      IF h^.besucht = TRUE THEN
         rek := TRUE
      ELSE
         h^.besucht := TRUE;

   h^.Parts^.gotoStart;
   { Elemente durcharbeiten, bis Listenende oder Rekursion. }
   WHILE (h^.Parts^.eol = FALSE) AND (rek = FALSE) DO BEGIN
      t := h^.Parts^.getData;

      { Wenn Rekursion vorhanden - rek TRUE setzen. }
      IF chkrek (t) = TRUE THEN
         rek := TRUE;
      h^.Parts^.gotoNext;
   END;

   h^.besucht := FALSE;

   { Bei gefundener Rekursion TRUE zurckliefern. }
   IF rek THEN
      chkrek := TRUE
   ELSE
      chkrek := FALSE;
END;




{ Funktion find.
  Sucht ein Element in der Adjazenzliste.
  @param  s Kurzbezeichnung des Elements.
  @return Pointer auf das gefundene Element.
          NIL wenn das Element nicht gefunden wurde. }
FUNCTION AdjazenzListe.find (s: string4) : AdjKnotenPtr;

VAR
   h, t: AdjKnotenPtr;

BEGIN
   adj^.gotostart;
   h := NIL;

   WHILE (adj^.eol = FALSE) AND (h = NIL) DO BEGIN
      t := adj^.getData;
      IF t^.kurzbez = s THEN
         h := (adj^.getData)
      ELSE
         adj^.gotoNext;
   END;

   find := h;
END;




{ Funktion get_alle_Knoten.
  bergibt Pointer auf den Listenanfang. }
FUNCTION AdjazenzListe.get_alle_Knoten : ListePtr;

BEGIN
   get_alle_Knoten := adj;
END;





{ Funktion isEmpty.
  berprft, ob Knoten vorhanden sind.
  @return TRUE  wenn keine Knoten gefunden wurden.
          FALSE wenn Knoten gefunden wurden. }
FUNCTION AdjazenzListe.isEmpty : boolean;
BEGIN
   IF adj^.isEmpty THEN
      isEmpty := TRUE
   ELSE
      isEMPTY := FALSE;
END;




{ Funktion linearisieren.
  Ordnet die Adjazenzliste so um, dass alle Abhngigkeiten erfllt sind.
  @param  bez   Der Arbeitsvorgang, der linearisiert werden soll.
                NIL, wenn der ganze Graph linearisiert werden soll.
  @param  h     Die linearisierte Hilfsliste.
  @return TRUE  Wenn die Liste linearisiert werden konnte.
          FALSE Wenn ein Fehler aufgetreten ist. }
FUNCTION AdjazenzListe.linearisieren (bez : string4; var h : ListePtr;
                                     var min, max: integer) : boolean;

VAR
   l       : ListePtr;
   k, m    : AdjKnotenPtr;
   ok, abh : boolean;

BEGIN
   ok  := TRUE;

   { Fall notwendig Hilfsliste initialisieren. }
   IF h = NIL THEN
      h := new (ListePtr, init);

   { berprfung, ob der ganze Graph linearisiert werden soll. }
   IF length (bez) = 0 THEN BEGIN
      l := ziel_Knoten;

      { Zielknoten durcharbeiten und in Liste eintragen. }
      l^.gotoStart;
      WHILE NOT l^.eol DO BEGIN
         k := l^.getData;
         k^.besucht := TRUE;
         h^.addElement (k);
         min := min + k^.min;
         max := max + k^.max;
         l^.gotoNext;
      END;

      { Zielknoten erneut durcharbeiten und Abhngigkeiten verfolgen. }
      l^.gotoStart;
      WHILE NOT l^.eol AND ok DO BEGIN
         k := l^.getData;

         { Rekursiver Aufruf. }
         ok := linearisieren (k^.kurzbez, h, min, max);

         l^.gotoNext;
      END;

   { Einen Knoten als Zielknoten verwenden. }
   END ELSE BEGIN
      k := find (bez);

      { Abbruch, wenn nicht gefunden. }
      IF k = NIL THEN
         exit;

      { Arbeitsvorgang Markieren und Eintragen, falls kein Zielknoten. }
      IF NOT k^.besucht THEN BEGIN
         k^.besucht := TRUE;
         h^.addElement (k);
         min := min + k^.min;
         max := max + k^.max;
      END;

      { Knoten untersuchen. }
      k^.Parts^.gotoStart;
      WHILE NOT k^.Parts^.eol AND ok DO BEGIN
         m := k^.Parts^.getData;

         { Wenn Vorgang noch nicht besucht, weiter linearisieren. }
         IF NOT m^.besucht THEN
            ok  := linearisieren (m^.kurzbez, h, min, max)

         { Alten Eintrag lschen, und hinten anhngen. }
         ELSE BEGIN
            ok := h^.find (m);
            h^.removeElement;
            h^.addElement (m);
            ok := linearisieren (m^.kurzbez, h, min, max);
         END;

         k^.Parts^.gotoNext;
      END;
   END;

   { Rckgabewert. }
   linearisieren := ok;
END;




{ Funktion remove_Knoten.
  Entfernt einen Knoten und alle seine Abhngigkeiten.
  @param  bez Kurzbezeichnung des zu entfernenden Knoten.
  @return TRUE  wenn der Knoten entfernt werden konnte.
          FALSE wenn der Knoten nicht entfernt werden konnte. }
FUNCTION AdjazenzListe.remove_Knoten (bez: string4) : boolean;

VAR
   h1, h2: AdjKnotenPtr;
   s     : string4;
   b, ok : boolean;

BEGIN
   h1 := find (bez);

   { PartOf Verknpfungen verfolgen und Abhngigkeiten entfernen. }
   h1^.partOf^.gotoStart;
   WHILE h1^.partOf^.eol = FALSE DO BEGIN
      h2 := h1^.partOf^.getData;
      s  := h2^.kurzbez;
      ok := remove_Link (s, bez);
      h1^.partOf^.gotoStart;
   END;

   { Parts Verknpfungen verfolgen und Abhngigkeiten entfernen. }
   h1^.Parts^.gotoStart;
   WHILE h1^.Parts^.eol = FALSE DO BEGIN
      h2 := h1^.Parts^.getData;
      s := h2^.kurzbez;
      ok := remove_Link (bez, s);
      h1^.Parts^.gotoStart;
   END;

   { Knoten aus Adjazenzliste und Speicher lschen. }
   h1 := find (bez);
   adj^.removeElement;
   dispose (h1);
END;




{ Funktino remove_Link.
  Entfernt eine Verknpfung zwischen zwei Knoten.
  @param  source Der bergeordnete Knoten, von dem die Verknpfung ausgeht.
  @param  dest   Der untergeordnete Knoten, in den die Verknpfung endet.
  @return TRUE  wenn die Verknfung gelscht werden konnte.
          FALSE wenn sie nicht gelscht werden konnte. }
FUNCTION AdjazenzListe.remove_Link (source, dest: string4) : boolean;

VAR
   h1, h2, t: AdjKnotenPtr;
   found    : boolean;

BEGIN
   found := FALSE;

   { Zu lschenden knoten suchen. }
   h1 := find (source);
   h2 := find (dest);

   { Wenn einer der Knoten nicht gefunden wurde - ABBRUCH. }
   IF (h1 = NIL) OR (h2 = NIL) THEN BEGIN
      remove_link := FALSE;
      exit;
   END;

   { partOf Verknpfung lschen, falls vorhanden. }
   h2^.partOf^.gotoStart;
   t := NIL;
   WHILE (h2^.partOf^.eol = FALSE) DO BEGIN
      t := h2^.partOf^.getData;
      IF t^.kurzbez = source THEN BEGIN
         h2^.partOf^.removeElement;
         h2^.partOf^.gotoEnd;
         h2^.partOf^.gotoNext;
         found := TRUE;
      END ELSE
         h2^.partOf^.gotoNext;
   END;

   { Parts Verknpfung lschen, falls vorhanden. }
   h1^.Parts^.gotoStart;
   t := NIL;
   WHILE (h1^.Parts^.eol = FALSE) DO BEGIN
      t := h1^.Parts^.getData;
      IF t^.kurzbez = dest THEN BEGIN
         h1^.Parts^.removeElement;
         h1^.Parts^.gotoEnd;
         h1^.Parts^.gotoNext;
      END ELSE
         h1^.Parts^.gotoNext;
   END;

   { Wurde eine Verknpfung entfernt? }
   IF NOT found THEN
      remove_link := FALSE
   ELSE
      remove_link := TRUE;
END;




{ Prozedur reset_Besucht.
  Setzt in allen Adjazenzknoten besucht auf FALSE. }
PROCEDURE AdjazenzListe.reset_Besucht;

VAR
   h: adjKnotenPtr;

BEGIN
   adj^.gotoStart;

   WHILE NOT adj^.eol DO BEGIN
      h := adj^.getData;
      h^.besucht := false;

      adj^.gotoNext;
   END;
END;




{ Funktion start_Knoten.
  Gibt alle Knoten als Liste zurck, die nicht von einem anderen
  Prozess abhngen.
  @return Liste mit allen Startknoten. }
FUNCTION AdjazenzListe.start_Knoten : ListePtr;

VAR
   h: ListePtr;
   t: AdjKnotenPtr;

BEGIN
   h := new (ListePtr, Init);

   { Adjazenzliste durcharbeiten. }
   adj^.gotoStart;
   WHILE adj^.eol = FALSE DO BEGIN
      t := adj^.getData;
      IF t^.Parts^.isEmpty = TRUE THEN
         h^.addElement (t);

      adj^.gotoNext;
   END;

   start_Knoten := h;
END;




{ Funktion ziel_Knoten.
  Gibt alle Zielknoten als Liste zurck.
  @return Liste mit allen Zielknoten. }
FUNCTION AdjazenzListe.ziel_Knoten : ListePtr;

VAR
   h: ListePtr;
   t: AdjKnotenPtr;

BEGIN
   h := new (ListePtr, Init);

   { Adjazenzliste durcharbeiten. }
   adj^.gotoStart;
   WHILE adj^.eol = FALSE DO BEGIN
      t := adj^.getData;
      IF t^.partOf^.isEmpty = TRUE THEN
         h^.addElement (t);

      adj^.gotoNext;
   END;

   ziel_Knoten := h;
END;




{ Procedure kritischer_Pfad.
  }
PROCEDURE AdjazenzListe.kritischer_Pfad (bez: string4; var max: integer;
                                         var krit: ListePtr);

VAR
   h, l: ListePtr;
   k, m: AdjKnotenPtr;
   max2: integer;

BEGIN
   { Fall notwendig kritische Liste initialisieren. }
   IF krit = NIL THEN
      krit := new (ListePtr, init);

   { berprfung, ob Zielknoten gesucht werden mssen. }
   IF length (bez) = 0 THEN BEGIN
      l := ziel_Knoten;

      { Zeilknoten erneut durcharbeiten und Abhngigkeiten verfolgen. }
      l^.gotoStart;
      WHILE NOT l^.eol DO BEGIN
         m    := l^.getData;
         max2 := 0;
         h    := NIL;

         { Rekursiver Aufruf. }
         kritischer_Pfad (m^.kurzbez, max2, h);

         { Wenn untersuchter Weg grsser als bisheriger grsster,
           dann den alten verwerfen. }
         IF max2 > max THEN BEGIN
            krit := h;
            max  := max2;
         END ELSE IF max2 = max THEN BEGIN
            h^.gotoStart;
            WHILE NOT h^.eol DO BEGIN
               IF NOT krit^.find (h^.getData) THEN
                  krit^.addElement (h^.getData);
               h^.gotoNext;
            END;
         END;

         l^.gotoNext;
      END;



   END ELSE BEGIN
      k := find (bez);

      k^.Parts^.gotoStart;
      WHILE NOT k^.Parts^.eol DO BEGIN
         m    := k^.Parts^.getData;
         max2 := 0;
         h    := NIL;

         { Rekursiver Aufruf. }
         kritischer_Pfad (m^.kurzbez, max2, h);

         { Wenn untersuchter Weg grsser als bisheriger grsster,
           dann den alten verwerfen. }
         IF max2 > max THEN BEGIN
            krit := h;
            max  := max2;
         END ELSE

         IF max2 = max THEN BEGIN
            h^.gotoStart;
            WHILE NOT h^.eol DO BEGIN
               IF NOT krit^.find (h^.getData) THEN
                  krit^.addElement (h^.getData);
               h^.gotoNext;
            END;
         END;

         k^.Parts^.gotoNext;
      END;

      max := max + k^.max;
      krit^.addElement (k);
   END;
END;





BEGIN
END.
